import time
import typing

from qiskit import BasicAer, execute
from qiskit.extensions import Initialize
from qiskit_textbook.tools import random_state

from .exceptions import CommunicationSystemException
from .instructions_set import CircuitInstruction
from .motherboard import Motherboard


class CommunicationSystem:
    def __init__(self, motherboard: Motherboard) -> None:
        self._information: typing.Any = random_state(1)
        self._motherboard: Motherboard = motherboard
        self._motherboard.circuit.append(Initialize(self._information), [0])
        self._motherboard.circuit.barrier()

    def _add_gate(self, instruction: CircuitInstruction) -> None:
        if instruction.type == "cnot":
            self._motherboard.circuit.cx(
                self._motherboard.quantum_registers[instruction.register_indexes[0]],
                self._motherboard.quantum_registers[instruction.register_indexes[1]],
            )
        elif instruction.type == "hadamard":
            self._motherboard.circuit.h(
                self._motherboard.quantum_registers[instruction.register_indexes[0]]
            )

    def measure_qubits(self) -> None:
        self._motherboard.circuit.measure(self._motherboard.quantum_registers[0], 0)
        self._motherboard.circuit.measure(self._motherboard.quantum_registers[1], 1)

    def decode(self) -> None:
        self._motherboard.circuit.x(2).c_if(self._motherboard.classical_registers[1], 1)
        self._motherboard.circuit.z(2).c_if(self._motherboard.classical_registers[0], 1)

    def test_output(self) -> bool:
        backend = BasicAer.get_backend("statevector_simulator")
        out_vector = (
            execute(self._motherboard.circuit, backend)
            .result()
            .get_statevector(decimals=6)
        )
        out_vector = [element for element in out_vector if element != 0]
        self._information = list(self._information.round(decimals=6))
        time.sleep(1)
        return self._information == out_vector

    def add_instructions(self, instructions: typing.List):
        if len(instructions) > 10:
            raise CommunicationSystemException("Instruction set is too big")
        [self._add_gate(CircuitInstruction(**(eval(gate)))) for gate in instructions]
